import sys
import collections
import yaml
import yschema


def validate_dictionaries(schema, documents, schema_name, doc_names):
    """
    Validate the given documents (list of dicts) against the schema (a
    dictionary). This function returns 0 on success and the number of
    failed documents on (partial) failure. Returns -1 if the schema is not
    valid (for some subset of schema problems).
    """
    print('YSchema is reading schema', schema_name)
    indent = '  '
    prefix1 = indent + 'ERROR: '
    prefix2 = ' ' * len(prefix1)
    try:
        schema = yschema.Schema(schema)
        print(indent + 'OK')
    except Exception as e:
        print(indent + 'Invalid schema!')
        print(indent + str(e).replace('\n', '\n' + indent))
        return -1

    num_errors = 0
    for doc, name in zip(documents, doc_names):
        print('Validating document', name)
        errors = schema.validate(doc)
        if errors:
            for e in errors:
                print(prefix1 + e.replace('\n', '\n' + prefix2))
            num_errors += 1
        else:
            print('  OK')
    return num_errors


def run_from_console():
    """
    Parse command line arguments and then run the validator
    """
    import argparse
    afile = argparse.FileType(mode='rt', encoding='utf8')
    parser = argparse.ArgumentParser(prog='YSchema',
                                     description='YAML schema validator')
    parser.add_argument('schemafile', help='Name of the file containing the '
                        'YSchema definition', type=afile)
    parser.add_argument('datafiles', help='Name of the file(s) containing data '
                        'to be validated', nargs='+', type=afile)
    args = parser.parse_args()

    # Ensure YAML reads data in an ordered way
    yaml.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                         lambda loader, node: collections.OrderedDict(loader.construct_pairs(node)))

    # Read the YAML files
    schema = yaml.safe_load(args.schemafile)
    documents = [yaml.safe_load(datafile) for datafile in args.datafiles]
    document_names = [datafile.name for datafile in args.datafiles]

    # Run the validation
    return validate_dictionaries(schema, documents, args.schemafile.name,
                                 document_names)


if __name__ == '__main__':
    sys.exit(run_from_console())



from .schema import Schema
from .exceptions import ValidationError, SchemaError
from .types import TYPES


__version__ = '1.0.1'


def validate(data, schema, types=None, constants=None, return_errors=False):
    """
    Given a dictionary with data and a dictionary with a schema, either
    raise a ValidationError if the document is invalid or return without
    an error for valid data. The schema is assumed to be valid.

    You can run str() on a :class:`ValidationError` to get an
    explaination for why the data did not validate. If return_errors
    is True then the validation will return a list of error messages
    instead of raising an exception. Schema errors will still raise.
    """
    def raise_error(message):
        if return_errors:
            return [message]
        else:
            raise ValidationError(message)

    if not isinstance(data, dict):
        raise_error('The data is not a dict, it is %s' % type(data))
    elif not isinstance(schema, dict):
        raise_error('The schema is not a dict, it is %s' % type(schema))

    all_types = TYPES.copy()
    if types is not None:
        all_types.update(types)

    all_consts = {}
    if constants is not None:
        all_consts.update(constants)

    schema = Schema(schema, all_types, all_consts)
    errors = schema.validate(data)

    if return_errors:
        return errors
    elif errors:
        raise ValidationError('\n'.join(errors))



from collections import Mapping
import fnmatch
import inspect
from .types import BaseType, TYPES
from .exceptions import ValidationError, SchemaError


class AliasedType(BaseType):
    def __init__(self, types, constants, typedef, name):
        """
        Make an alias to a defined type. The alias type is created with
        parameter initialization on alias definition time an cannot take
        more parameters when the alias is used
        """
        self.name = name
        self.aliased_type = _make_type(types, constants, typedef, name)

    def validate(self, data, key):
        return self.aliased_type.validate(data, key)

    def __repr__(self):
        return '<Alias %s of %r>' % (self.name, self.aliased_type)


class Schema(BaseType):
    def __init__(self, schema_data, types=None, constants=None, name='__root__',
                 config=None):
        """
        This is a YSchema parser and validator. The schema is parsed
        when instantiating this class, raising :exception:`SchemaError`
        on errors. This is a subclass of the more general exception
        :exception:`ValidationError`.

        The schema_data describing the schema should be a dictionary.
        After creating this class call :method:`validate`
        """
        self.type_name = name
        self.types = types if types else TYPES.copy()
        self.constants = constants if constants else {}

        self.keys = {}
        self.glob_keys = []
        self.required_keys = set()
        self.config = {}

        # Skip keys starting with a given prefix (default: skip no keys)
        self.config['skip_key_prefix'] = ''

        if config:
            self.config.update(config)

        self._setup_schema(schema_data)

    def _setup_schema(self, schema):
        """
        Parse the schema and any sub schema (defined types) given in the
        schema dictionary
        """
        #  If there is an empty (sub) schema PyYAML will return None
        if schema is None:
            schema = {}

        for key, value in schema.items():
            if key == 'inherit':
                super_type = _make_type(self.types, self.constants,
                                        value, 'super(%s)' % self.type_name)
                if not isinstance(super_type, Schema):
                    raise SchemaError('Invalid super type %r' % super_type)
                self.types.update(super_type.types)
                self.keys.update(super_type.keys)
                self.required_keys.update(super_type.required_keys)
                self.config.update(super_type.config)
                self.glob_keys.extend(super_type.glob_keys)
                continue

            cmd, name = key.split()

            if cmd == 'constant':
                self.constants[name] = value
            elif cmd == 'type':
                consts = self.constants.copy()
                types = self.types.copy()
                config = self.config.copy()
                self.types[name] = Schema(value, types, consts, name=name,
                                          config=config)
            elif cmd == 'alias':
                self.types[name] = AliasedType(self.types, self.constants,
                                               value, name)
            elif cmd == 'required':
                self.required_keys.add(name)
                if '*' in name:
                    raise SchemaError('Cannot use glob for required parameter'
                                      ' %r' % name)
                self.keys[name] = _make_type(self.types, self.constants,
                                             value, name)
            elif cmd == 'optional':
                if name != '*' and '*' in name:
                    # Handle glob keys except the glob-all key
                    self.glob_keys.append(name)
                self.keys[name] = _make_type(self.types, self.constants,
                                             value, name)
            elif cmd == 'config':
                if name in self.config:
                    self.config[name] = value
                else:
                    raise SchemaError('Unknown schema config %r' % name)
            else:
                raise SchemaError('Unknown schema command %r' % cmd)

    def validate(self, data, key='__root__'):
        """
        Validate the dictionary ``data`` according to this schema. This
        method is recursive and will call itself on sub-dictionaries
        (``self`` will then be a sub-schema). For other value types this
        method will call validate() on one of the types defined in
        :mod:`yschema.types` (or a user defined type). The method returns
        a non-empty list of error messages if the document is not valid
        according to the schema, otherwise an empty list
        """
        errors = []
        prefix = ''
        if key != '__root__':
            # prefix = '%s %s: ' % (self.type_name, key)
            prefix = '%s: ' % (key, )

        if not isinstance(data, Mapping):
            errors.append('%sa (sub) dictionary must be a mapping, but %r '
                          'got a %s' % (prefix, key, type(data)))
            return errors

        seen_keys = set()
        skip_prefix = self.config['skip_key_prefix']
        for subkey, value in data.items():
            if skip_prefix and subkey.startswith(skip_prefix):
                continue
            seen_keys.add(subkey)

            # The key must be registered, or a '*' key must be present
            val_type = None
            if subkey in self.keys:
                val_type = self.keys[subkey]
            else:
                for pattern in self.glob_keys:
                    if fnmatch.fnmatch(subkey, pattern):
                        val_type = self.keys[pattern]
                        break
                else:
                    # Match everything last to be most specific first
                    if '*' in self.keys:
                        val_type = self.keys['*']

            if val_type is None:
                errors.append('%sgot unexpected key %r' % (prefix, subkey))
                continue

            # Validate this key: value pair
            for err in val_type.validate(value, subkey):
                errors.append('%s%s' % (prefix, err))

        diff = self.required_keys.difference(seen_keys)
        if diff:
            errors.append('%sMissing required keys %r' % (prefix, diff))

        return errors


def _get_type(types, type_name, key, params):
    """
    Get a type by name and instantiate it, or raise an appropriate error
    """
    if type_name in types:
        t = types[type_name]
        if inspect.isclass(t):
            return t(**params)

        if params:
            raise SchemaError('Cannot give further parameters to %r in key %r'
                              % (t.type_name, key))
        return t
    else:
        raise SchemaError('The referenced type %r is not defined' % type_name)


def _make_type(types, constants, typedef, key):
    """
    Get an instantiated type class.

    The typedef string can either be just a name 'str', 'int' etc or
    a function call 'int(min_val=2)'. The last one makes the code eval
    "dict(min_val=2)" to get the parameters. In this evaluation the
    given constants dictionary is used.

    The type name must be defined in the types dictionary. The key is
    included only to return better error messages in the exception text.
    """
    i = typedef.find('(')
    if i == -1:
        # Definition without any arguments
        return _get_type(types, typedef.strip(), key, {})

    # The definition includes some arguments
    type_name = typedef[:i].strip()
    type_args = typedef[i:]
    if '=' not in type_args:
        raise SchemaError('Missing key=value format in type for %r, got %r'
                          % (key, typedef))
    code = 'dict%s' % type_args
    local_vars = constants.copy()
    local_vars.update(types)
    try:
        args = eval(code, globals(), local_vars)
    except Exception as e:
        raise SchemaError('Could not define type for %r, got %s'
                          % (key, e))
    return _get_type(types, type_name, key, args)



"""
YSchema Types that are used when validating

All built in validation types except (sub) schema and alias are defined
below. The built in types are registered in the TYPES dictionary. Any
type used in a schema must be either defined in TYPES, defined as an
alias in the schema, or a user provided type given when parsing the
schema.

A Type is an object that implements a validate() method that takes value
and key name and returns a list of error messages (an empty list for valid
data). Only the value is typically validated, the key is used to construct
error messages.
"""
import inspect


PINF = float('+inf')
MINF = float('-inf')
TYPES = {}


def register_schema_type(name):
    """
    A class decorator to register YSchema types
    """
    def register(schema_type_class):
        TYPES[name] = schema_type_class
        schema_type_class.type_name = name
        return schema_type_class
    return register


def get_type_object(candidate):
    """
    To allow users to specify things like ``any_of(types=(str, float))``
    or ``any_of(types=(str(min_len=4), float))``, this function checks
    for instance or class and instantiates if necessary
    """
    if inspect.isclass(candidate):
        candidate = candidate()
    return candidate


class BaseType:
    def __repr__(self):
        return '<SchemaType %s>' % self.type_name


@register_schema_type('Any')
class AnyType(BaseType):
    def validate(self, value, key):
        return []


@register_schema_type('bool')
class BoolType(BaseType):
    def __init__(self, equals=None):
        self.equals = equals

    def validate(self, value, key):
        if not isinstance(value, bool):
            return ['Expected a boolean for key %r, got %r which'
                    ' is a %s' % (key, value, type(value))]

        errors = []
        if self.equals is not None:
            if value != self.equals:
                errors.append('Expected %r == %r, got %r' %
                              (key, self.equals, value))
        return errors


@register_schema_type('str')
class StringType(BaseType):
    def __init__(self, min_len=0, max_len=1e100, equals=None, prefix=None):
        self.min_len = min_len
        self.max_len = max_len
        self.equals = equals
        self.prefix = prefix

    def validate(self, value, key):
        if not isinstance(value, str):
            return ['Expected a string for key %r, got %r which '
                    'is a %s' % (key, value, type(value))]

        errors = []
        if len(value) < self.min_len:
            errors.append('Expected a string of length at least %r for'
                          ' %r, got %r which has length %r' %
                          (self.min_len, key, value, len(value)))

        if len(value) > self.max_len:
            errors.append('Expected a string of length at most %r for'
                          ' %r, got %r which has length %r' %
                          (self.max_len, key, value, len(value)))

        if isinstance(self.equals, str):
            if value != self.equals:
                errors.append('Expected %r = %r, got %r' %
                              (key, self.equals, value))
        elif self.equals is not None:
            if value not in self.equals:
                errors.append('Expected %r in %r, got %r' %
                              (key, self.equals, value))

        if self.prefix is not None:
            if not value.startswith(self.prefix):
                errors.append('Expected %r starting with %r, got %r' %
                              (key, self.prefix, value))
        return errors


@register_schema_type('int')
class IntType(BaseType):
    def __init__(self, min_val=MINF, max_val=PINF, equals=None):
        self.min_val = min_val
        self.max_val = max_val
        self.equals = equals

    def validate(self, value, key):
        if not isinstance(value, int):
            return ['Expected an integer for key %r, got %r which'
                    ' is a %s' % (key, value, type(value))]

        errors = []
        if value < self.min_val:
            errors.append('Expected an integer larger than %r for '
                          '%r, got %r' % (self.min_val, key, value))

        if value > self.max_val:
            errors.append('Expected an integer smaller than %r for '
                          '%r, got %r' % (self.max_val, key, value))

        if isinstance(self.equals, int):
            if value != self.equals:
                errors.append('Expected %r = %r, got %r' %
                              (key, self.equals, value))
        elif self.equals is not None:
            if value not in self.equals:
                errors.append('Expected %r in %r, got %r' %
                              (key, self.equals, value))
        return errors


@register_schema_type('float')
class FloatType(BaseType):
    def __init__(self, min_val=MINF, max_val=PINF, equals=None):
        self.min_val = min_val
        self.max_val = max_val
        self.equals = equals

    def validate(self, value, key):
        num = (float, int)
        if not isinstance(value, num):
            return ['Expected a float for key %r, got %r which'
                    ' is a %s' % (key, value, type(value))]

        errors = []
        if value < self.min_val:
            errors.append('Expected a float larger than %r for '
                          '%r, got %r' % (self.min_val, key, value))

        if value > self.max_val:
            errors.append('Expected a float smaller than %r for '
                          '%r, got %r' % (self.max_val, key, value))

        if isinstance(self.equals, num):
            if value != self.equals:
                errors.append('Expected %r = %r, got %r' %
                              (key, self.equals, value))
        elif self.equals is not None:
            if value not in self.equals:
                errors.append('Expected %r in %r, got %r' %
                              (key, self.equals, value))
        return errors


@register_schema_type('list')
class ListType(BaseType):
    def __init__(self, type=None, min_len=0, max_len=1e100):
        self.item_type = type
        self.min_len = min_len
        self.max_len = max_len

    def validate(self, value, key):
        if not isinstance(value, list):
            return ['Expected a list for key %r, got %r which '
                    'is a %s' % (key, value, type(value))]

        errors = []
        if len(value) < self.min_len:
            errors.append('Expected a list of length at least %r for'
                          ' %r, got %r which has length %r' %
                          (self.min_len, key, value, len(value)))

        if len(value) > self.max_len:
            errors.append('Expected a list of length at most %r for'
                          ' %r, got %r which has length %r' %
                          (self.max_len, key, value, len(value)))

        if self.item_type is not None:
            required_type = get_type_object(self.item_type)
            for i, item in enumerate(value):
                errors.extend(required_type.validate(item, '%s[%d]' % (key, i)))
        return errors


@register_schema_type('one_of')
class OneOfType(BaseType):
    def __init__(self, types):
        self.types = types

    def validate(self, value, key):
        valid = set()
        invalid = set()

        reasons = []
        for possible_type in self.types:
            possible_type = get_type_object(possible_type)
            errs = possible_type.validate(value, key)
            if not errs:
                valid.add(possible_type)
            else:
                reasons.append('\n'.join(errs).replace('\n', '\n    '))
                invalid.add(possible_type)
        reasons = '. Tried:\n' + '\n'.join('  - %s' % r for r in reasons)

        if len(valid) > 1:
            return ['Expected %r to be exactly one type,'
                    ' not %d types %r'
                    % (key, len(valid), valid) + reasons]
        elif not valid:
            return ['%r is not one of %r' % (key, invalid) + reasons]
        else:
            return []


@register_schema_type('any_of')
class AnyOfType(BaseType):
    def __init__(self, types):
        self.types = types

    def validate(self, value, key):
        valid = set()
        invalid = set()

        reasons = []
        for possible_type in self.types:
            possible_type = get_type_object(possible_type)
            errs = possible_type.validate(value, key)
            if not errs:
                valid.add(possible_type)
            else:
                reasons.append('\n'.join(errs).replace('\n', '\n    '))
                invalid.add(possible_type)
        reasons = '. Tried:\n' + '\n'.join('  - %s' % r for r in reasons)

        if not valid:
            return ['%r is not one of %r' % (key, invalid) + reasons]
        else:
            return []



from setuptools import setup, find_packages
from codecs import open
import os


here = os.path.abspath(os.path.dirname(__file__))


# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


# Get the version
for line in open(os.path.join(here, 'yschema', '__init__.py')):
    if line.startswith('__version__'):
        version = line.split('=')[1].strip()[1:-1]


# Give setuptools/pip informattion about the Ocellaris package
setup(
    name='yschema',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,

    description='A YAML/JSON/dictionary schema validator with terse schema definitions',
    long_description=long_description,

    # The project's main homepage.
    url='https://bitbucket.org/trlandet/yschema',

    # Author details
    author='Tormod Landet',
    author_email='tormod@landet.net',

    # Choose your license
    license='Apache 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Software Development :: Quality Assurance',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
    ],

    # What does your project relate to?
    keywords='yaml json schema validation',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed.
    install_requires=['PyYAML'],

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={},
    zip_safe=True,

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'yschema=yschema.__main__:run_from_console',
        ],
    }
)
